"use strict";
/******************************************************************************
 * Copyright 2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.init = exports.inject = void 0;
// used to tag groups and services to be eagerly evaluated
const isEager = Symbol();
// used to tag proxy properties during initialization
const isRequested = Symbol();
/**
 * Creates a container from the given modules.
 * The modules are merged and the dependencies are injected.
 * The container is a proxy that lazily creates the dependencies.
 * The dependencies are cached and reused.
 *
 * @param modules The modules to be merged and injected.
 * @returns The container.
 */
// @ts-expect-error A rest parameter must be of an array type. ts(2370)
function inject(...modules) {
    const module = modules.reduce(merge, {});
    const container = proxify(module);
    initializeEagerServices(module, container);
    return container;
}
exports.inject = inject;
function merge(target, source) {
    Object.keys(source).forEach(key => {
        target[key] = isObj(source[key]) && isObj(target[key]) ? merge(target[key], source[key]) : source[key];
    });
    return target;
}
function isObj(t) {
    return t !== null && typeof t === 'object';
}
/**
 * Marks a group or factory to be eagerly evaluated.
 * Eager groups and factories are evaluated when the container is created.
 *
 * @param val The group or factory to be eagerly evaluated.
 * @returns A new instance of the group or factory.
 */
function init(val) {
    return (isEager in val) ? val : Object.assign((typeof val === 'function') ? ((...args) => val(args)) : Object.assign({}, val), { [isEager]: true });
}
exports.init = init;
function initializeEagerServices(val, context, eagerGroup = false) {
    const eagerScope = eagerGroup || isEager in val;
    (typeof val === 'function') ? eagerScope && val(context) : Object.keys(val).forEach(key => {
        initializeEagerServices(val[key], context, eagerScope);
    });
}
function proxify(module, container, path = '') {
    const get = (obj, prop, proxy) => {
        const name = path + '[' + String(prop) + ']';
        if (obj[prop] === isRequested) {
            throw new Error('Cyclic dependency ' + name + '. See https://djinject.io/#cyclic-dependencies');
        }
        const ctr = container || proxy;
        if (prop in obj) {
            return obj[prop];
        }
        else if (prop in module) {
            obj[prop] = isRequested;
            const val = module[prop];
            return obj[prop] = (typeof val === 'function') ? val(ctr) : proxify(val, ctr, name);
        }
        else {
            return undefined;
        }
    };
    const proxy = new Proxy({}, {
        deleteProperty: () => false,
        get,
        getOwnPropertyDescriptor: (target, prop) => (get(target, prop, proxy), Object.getOwnPropertyDescriptor(target, prop)),
        has: (_, prop) => prop in module,
        ownKeys: () => Reflect.ownKeys(module)
    });
    return proxy;
}
//# sourceMappingURL=inject.js.map