/******************************************************************************
 * Copyright 2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
/**
 * Type of an indexed structure, excluding interfaces, classes, arrays and functions.
 * Symbol keys are intentionally not supported.
 */
export type Obj = Record<string | number, unknown>;
/**
 * A Module is an Obj type consisting of nested groups and dependency factories.
 * A dependency factory is a function that creates a dependency given a context.
 * The module does not need to fully cover the context.
 */
export type Module<Context extends Obj = any, Dependencies extends Obj = {}> = IsUniversal<Dependencies> extends true ? Dependencies : Prettify<Merge<RequiredModule<Merge<Dependencies, Context>, Dependencies>, PartialModule<Merge<Dependencies, Context>, Context>>>;
type PartialModule<Context, Dependencies> = {
    [K in keyof Dependencies]?: (Dependencies[K] extends Obj | undefined ? PartialModule<Context, NoUndefined<Dependencies[K]>> : never) | Factory<Context, NoUndefined<Dependencies[K]>>;
};
type RequiredModule<Context, Dependencies> = {
    [K in keyof Dependencies]: (Dependencies[K] extends Obj | undefined ? RequiredModule<Context, NoUndefined<Dependencies[K]>> : never) | Factory<Context, NoUndefined<Dependencies[K]>>;
};
/**
 * A function that creates a Dependency given a Context.
 */
export type Factory<Context, Dependency> = (ctx: Context) => Dependency;
/**
 * The type of an inversion of control container has the same structure as the merged
 * Modules, but with the dependency factories replaced by the dependencies.
 */
export type Container<Modules extends Module[]> = Prettify<_Container<Modules>>;
type _Container<Modules extends Module[], Dependencies extends Obj = ReflectDependencies<Modules>, CheckResult = Check<Modules>> = CheckResult extends DjinjectError ? CheckResult : Dependencies;
/**
 * Checks the Modules for conflicts and missing dependencies.
 */
export type Check<Modules extends Module[]> = _Check<Modules>;
type _Check<Modules extends Module[], Context extends Obj = ReflectContext<Modules>, Dependencies extends Obj = ReflectDependencies<Modules>, ContextConflicts extends any[] = CheckContextConflicts<Context, Dependencies>, DependencyConflicts extends any[] = CheckDependencyConflicts<Dependencies>, MissingDependencies extends any[] = CheckMissingDependencies<Context, Dependencies>, Errors = [
    ...ContextConflicts,
    ...DependencyConflicts,
    ...MissingDependencies
]> = Errors extends [] ? Modules : DjinjectError<{
    context_conflicts: ContextConflicts;
    dependency_conflicts: DependencyConflicts;
    missing_dependencies: MissingDependencies;
}>;
declare const djinject_error: unique symbol;
export type DjinjectError<Errors extends Record<string, string | string[]> = any> = Prettify<{
    [djinject_error]: FilterNot<Errors, []>;
}>;
type ReflectContext<Modules, Module = ReplaceFactories<Modules>> = MergeArray<UnionToTuple<InferContext<Module[keyof Module]>>>;
type ReplaceFactories<Modules> = Modules extends [infer HEAD, ...infer TAIL] ? Replace<ReplaceFactories<TAIL>, Paths<HEAD>> : {};
type InferContext<F> = F extends Factory<infer Context, any> ? Context & {} : never;
type ReflectDependencies<Modules> = Modules extends [infer HEAD, ...infer TAIL] ? Merge<ReflectDependencies<TAIL>, InferDependencies<HEAD>> : {};
type InferDependencies<Module> = Module extends Factory<any, infer Dependency> ? Dependency : IsUniversal<Module> extends true ? Module : {
    [K in keyof Module]: InferDependencies<Module[K]>;
};
type CheckContextConflicts<Context, Dependencies, Props = Filter<Paths<Context> & FilterNot<Paths<Dependencies>, never>, never>> = IsEmpty<Props> extends true ? [] : UnionToTuple<keyof Props>;
type CheckDependencyConflicts<Dependencies, Props = Filter<Paths<Dependencies>, never>> = IsEmpty<Props> extends true ? [] : UnionToTuple<keyof Props>;
type CheckMissingDependencies<Context, Dependencies, Props = Omit<Paths<Context>, keyof Paths<Dependencies>>> = true extends (IsUniversal<Context> | IsEmpty<Props>) ? [] : UnionToTuple<keyof Props>;
type Paths<T> = true extends (IsUniversal<T> | IsNotObj<T> | IsEmpty<T>) ? T : Prettify<UnionToIntersection<TupledPaths<DeepRequired<T>> extends infer P ? P extends [string, unknown] ? {
    [K in P[0]]: P[1];
} : never : never>>;
type TupledPaths<T, K = keyof T> = K extends Exclude<keyof T, symbol> ? true extends (IsEmpty<T[K]> | IsUniversal<T[K]> | IsNotObj<T[K]>) ? [`${K}`, T[K]] : TupledPaths<T[K]> extends infer F ? F extends [string, any] ? [`${K}.${F[0]}`, F[1]] : never : never : never;
type DeepRequired<T> = T extends Obj ? {
    [K in keyof T]-?: DeepRequired<T[K]>;
} : T;
type MergeArray<A> = A extends [infer HEAD, ...infer TAIL] ? Merge<MergeArray<TAIL>, HEAD> : {};
type Merge<Source, Target> = IsUniversal<Source> extends true ? Source : IsUniversal<Target> extends true ? Target : Prettify<Omit<Target, keyof Source> & {
    [K in keyof Source]: K extends keyof Target ? Source[K] extends Obj ? Target[K] extends Obj ? Merge<Source[K], Target[K]> : never : Target[K] extends Obj ? never : (Source[K] extends Target[K] ? Source[K] : never) : Source[K];
}>;
type Replace<Source, Target> = IsUniversal<Source> extends true ? Source : IsUniversal<Target> extends true ? Target : Omit<Target, keyof Source> & {
    [K in keyof Source]: K extends keyof Target ? Source[K] extends Obj ? Target[K] extends Obj ? Replace<Source[K], Target[K]> : never : Target[K] extends Obj ? never : Source[K] : Source[K];
};
type Filter<T, V> = Pick<T, {
    [K in keyof T]-?: [T[K]] extends [V] ? K : never;
}[keyof T]>;
type FilterNot<T, V> = IsUniversal<T> extends true ? T : Omit<T, keyof Filter<T, V>>;
type IsEmpty<T> = T extends Obj ? [keyof T] extends [never] ? true : false : false;
type IsNotObj<T> = T extends Obj ? false : true;
type IsUniversal<T> = 0 extends 1 & T ? true : [T] extends [never] ? true : false;
type NoUndefined<T> = T extends undefined ? never : T;
type Prettify<T> = T extends Obj ? {
    [K in keyof T]: Prettify<T[K]>;
} : T;
type UnionToIntersection<U> = (U extends any ? (arg: U) => void : never) extends ((arg: infer I) => void) ? I : never;
type UnionToTuple<T> = [
    T
] extends [never] ? [] : [...UnionToTuple<Exclude<T, LastOf<T>>>, LastOf<T>];
type LastOf<T> = UnionToIntersection<T extends any ? () => T : never> extends () => infer R ? R : never;
export {};
//# sourceMappingURL=types.d.ts.map